import random

from .camera import Camera
from .chapter_transition import ChapterTransition
from .inventory import Inventory
from .player import Player
from .room import Room

from os import listdir
from os.path import isfile, join


class GameState(object):
    RESTART_QUIPS = [
        'It\'s only a mistake if you do it twice!',
        'Better luck next time!',
        'Live and learn',
        'Try, try again'
    ]

    def __init__(self):
        self.fps = 60
        self.scale = 2
        self.camera = Camera()
        self.rooms = {}
        self.room = None
        self.last_chapter_start_name = None
        self.player = Player((0, 0), self)
        self.time_period = 0  # 0, 1, or 2
        self.screen_size = (640, 480)
        self.inventory = Inventory(self)
        self.paused = False

        self.transition_queue = []
        self.editor_editing = False
        self.editor_tile_behavior_idx = 0
        self.editor_tileset_idx = 0
        self.editor_tile_idx = 0
        self.editor_layer_idx = 0

        self.load_room('0-1', True)

    def reset_level(self):
        self.rooms = {}
        self.is_drunk = False
        self.is_hyper = False
        self.is_pee = False
        self.is_tax = False
        self.is_bucket = False

    def offset_time_period(self, ofs):
        self.time_period = (self.time_period + ofs) % 3

    def update(self, dt):
        if self.transition_queue:
            self.transition_queue[-1].update(dt)
        else:
            self.room.update(dt)
            self.player.update(dt)

    def load_room(self, name, is_resetting=False):
        if is_resetting:
            self.reset_level()
            self.last_chapter_start_name = name

        # Keep levels in memory so any changes within them remain
        if name not in self.rooms:
            room = Room(name, self)
            room.reload(is_resetting)
            self.rooms[name] = room
        self.room = self.rooms[name]

    def restart_chapter(self):
        self.transition_queue.insert(
            0,
            ChapterTransition(self, random.choice(self.RESTART_QUIPS))
        )
        self.load_room(self.last_chapter_start_name, True)

    def load_prev_room(self):
        room_name = self.room.name
        level_files = [f for f in listdir('maps') if '.json' in f]
        level_files.sort()
        map_index = -1
        idx = 0
        for level_file in level_files:
            if room_name in level_file:
                map_index = idx
                break
            idx += 1
        if map_index > 0:
            self.load_room(level_files[map_index-1].strip('.json'), False)

    def load_next_room(self):
        room_name = self.room.name
        level_files = [f for f in listdir('maps') if '.json' in f]
        level_files.sort()
        map_index = -1
        idx = 0
        for level_file in level_files:
            if room_name in level_file:
                map_index = idx
                break
            idx += 1
        if map_index < len(level_files)-1:
            self.load_room(level_files[map_index+1].strip('.json'), False)

gamestate = GameState()