import pygame

from .baseobjects import Text
from .gamestate import gamestate
from .tiles import Tile, TILESETS, TILE_BEHAVIORS

class GameWindow(object):
    def __init__(self, title):
        self._screen_internal = pygame.Surface(gamestate.screen_size)
        self._screen_scaled = pygame.display.set_mode(
            (gamestate.screen_size[0] * gamestate.scale, gamestate.screen_size[1] * gamestate.scale),
            pygame.DOUBLEBUF
        )
        pygame.display.set_caption(title)

    def do_events(self, dt):
        mouse_x, mouse_y = pygame.mouse.get_pos()
        mouse_left, mouse_middle, mouse_right = pygame.mouse.get_pressed()
        keys = pygame.key.get_pressed()

        if gamestate.editor_editing:
            # Draw cursor for level editor
            mouse_tile_pos = [
                (((mouse_x // gamestate.scale) + gamestate.camera.position[0]) // Tile.SIZE) * Tile.SIZE,
                (((mouse_y // gamestate.scale) + gamestate.camera.position[1]) // Tile.SIZE) * Tile.SIZE
            ]

            # Cursor
            tileset_names = list(TILESETS.keys())
            tileset_name = tileset_names[gamestate.editor_tileset_idx % len(tileset_names)]

            behavior_names = list(TILE_BEHAVIORS.keys())
            behavior_name = behavior_names[gamestate.editor_tile_behavior_idx % len(behavior_names)]
            behavior = TILE_BEHAVIORS[behavior_name]

            editor_cursor = behavior(tileset_name, mouse_tile_pos, gamestate.editor_layer_idx, gamestate, gamestate.editor_tile_idx)
            self.draw(editor_cursor)

            # Debug text
            debug_font_size = 12 * gamestate.scale
            self.draw(Text((0, 0 * debug_font_size), gamestate, 'Room name: %s' % gamestate.room.name, debug_font_size))
            self.draw(Text((0, 1 * debug_font_size), gamestate, 'Tile layer (1/2): %d' % gamestate.editor_layer_idx, debug_font_size))
            self.draw(Text((0, 2 * debug_font_size), gamestate, 'Tileset (F2/F3): %s' % tileset_name, debug_font_size))
            self.draw(Text((0, 3 * debug_font_size), gamestate, 'Tile index (F4/F5): %d' % gamestate.editor_tile_idx, debug_font_size))
            self.draw(Text((0, 4 * debug_font_size), gamestate, 'Tile behavior(F6/F7): %s' % behavior_name, debug_font_size))
            self.draw(Text((0, 5 * debug_font_size), gamestate, 'Time period (F8/F9): %d' % gamestate.time_period, debug_font_size))

            hovered_tile_bg = gamestate.room.background.get(tuple(mouse_tile_pos), None)
            hovered_tile_ob = gamestate.room.obstacles.get(tuple(mouse_tile_pos), None)
            tile_info = ''
            if hovered_tile_bg is not None:
                tile_info += 'BG:' + hovered_tile_bg.behavior
            if hovered_tile_ob is not None:
                tile_info += ', OB:' + hovered_tile_ob.behavior
            self.draw(Text((0, 6 * debug_font_size), gamestate, 'Cursor: %s %s' % (str(mouse_tile_pos), tile_info), debug_font_size))

            if mouse_left or mouse_right:
                gamestate.room.remove_tile(editor_cursor.position, gamestate.editor_layer_idx)
                if mouse_left:
                    gamestate.room.add_tile(editor_cursor, gamestate.editor_layer_idx)

            # Free camera movement
            cam_ofs = gamestate.scale * 0.25 * dt
            if keys[pygame.K_LEFT]:
                gamestate.camera.apply_offset((-cam_ofs, 0))
            elif keys[pygame.K_RIGHT]:
                gamestate.camera.apply_offset((cam_ofs, 0))
            elif keys[pygame.K_UP]:
                gamestate.camera.apply_offset((0, -cam_ofs))
            elif keys[pygame.K_DOWN]:
                gamestate.camera.apply_offset((0, cam_ofs))

        elif not gamestate.paused and not gamestate.transition_queue:
            if keys[pygame.K_LEFT]:
                gamestate.player.move_direction((-1, 0))
            elif keys[pygame.K_RIGHT]:
                gamestate.player.move_direction((1, 0))
            elif keys[pygame.K_UP]:
                gamestate.player.move_direction((0, -1))
            elif keys[pygame.K_DOWN]:
                gamestate.player.move_direction((0, 1))

        # Handle events
        for evt in pygame.event.get():
            if evt.type == pygame.QUIT:
                return False
            elif evt.type == pygame.KEYDOWN:
                if evt.key == pygame.K_F1:
                    gamestate.editor_editing = not gamestate.editor_editing

                # Editor controls
                if gamestate.editor_editing:
                    if evt.key == pygame.K_F2:
                        gamestate.editor_tileset_idx -= 1
                    elif evt.key == pygame.K_F3:
                        gamestate.editor_tileset_idx += 1
                    elif evt.key == pygame.K_F4:
                        gamestate.editor_tile_idx -= 1
                    elif evt.key == pygame.K_F5:
                        gamestate.editor_tile_idx += 1
                    elif evt.key == pygame.K_F6:
                        gamestate.editor_tile_behavior_idx -= 1
                    elif evt.key == pygame.K_F7:
                        gamestate.editor_tile_behavior_idx += 1
                    elif evt.key == pygame.K_F8:
                        gamestate.offset_time_period(-1)
                    elif evt.key == pygame.K_F9:
                        gamestate.offset_time_period(1)
                    elif keys[pygame.K_LCTRL] and evt.key == pygame.K_s:
                        gamestate.room.save()
                    elif evt.key == pygame.K_r:
                        gamestate.load_room(gamestate.room.name, True)
                    elif evt.key == pygame.K_1:
                        gamestate.editor_layer_idx = 0
                    elif evt.key == pygame.K_2:
                        gamestate.editor_layer_idx = 1
                    elif evt.key == pygame.K_q:
                        return False
                    elif evt.key == pygame.K_o:
                        gamestate.load_prev_room()
                    elif evt.key == pygame.K_p:
                        gamestate.load_next_room()
                else:
                    if not gamestate.paused and not gamestate.transition_queue:
                        if keys[pygame.K_LCTRL] and evt.key == pygame.K_r:
                            gamestate.restart_chapter()
                    if keys[pygame.K_LCTRL] and evt.key == pygame.K_q:
                        return False
                    elif evt.key == pygame.K_ESCAPE:
                        gamestate.paused = not gamestate.paused

        return True
    
    def clear(self):
        self._screen_internal.fill((0, 0, 0))
    
    def draw(self, obj):
        obj.draw_on(self._screen_internal, gamestate.camera)
        pygame.transform.scale(self._screen_internal, self._screen_scaled.get_size(), self._screen_scaled)
