import json
import os

from .tiles import Tile, TILE_BEHAVIORS

class Room(object):
    def __init__(self, name, gamestate):
        self.name = name
        self.background = {}
        self.obstacles = {}
        self.gamestate = gamestate
        self.layers = []
        self.starting_time_period = None
        self.player_start_position = None
        self.camera_start_position = None
    
    def _draw_layer(self, layer, surface, cam):
        v = cam.get_view_rect(surface)
        for tile in layer.values():
            if v.contains(tile.get_rect()) or \
               v.colliderect(tile.get_rect()):
                tile.draw_on(surface, cam)

    def draw_on(self, surface, cam):
        self._draw_layer(self.background, surface, cam)
        self._draw_layer(self.obstacles, surface, cam)

    def add_tile(self, tile, layer_idx):
        self.layers[layer_idx][tuple(tile.position)] = tile
        tile.room = self

    def remove_tile(self, position, layer_idx):
        pos = tuple(position)
        layer = self.layers[layer_idx]
        if pos in layer:
            del layer[pos]

    def as_dict(self):
        d = {
            'background': [t.as_dict() for t in self.background.values()],
            'obstacles': [t.as_dict() for t in self.obstacles.values()],
        }
        if self.starting_time_period is not None:
            d['starting_time_period'] = self.starting_time_period
        if self.player_start_position is not None:
            d['player_start_position'] = self.player_start_position
        if self.camera_start_position is not None:
            d['camera_start_position'] = self.camera_start_position
        return d

    def _load_tile(self, tile_dict, layer_idx):
        behavior = TILE_BEHAVIORS[tile_dict.get('behavior', 'tile')]
        tile_dict['gamestate'] = self.gamestate
        tile_dict['layer_index'] = layer_idx
        del tile_dict['behavior']
        self.add_tile(behavior(**tile_dict), layer_idx)

    def get_path(self):
        return os.path.join('maps', self.name + '.json')

    def reload(self, is_resetting):
        self.background = {}
        self.obstacles = {}
        self.layers = [
            self.background,
            self.obstacles
        ]
        with open(self.get_path(), 'r') as f:
            j = json.load(f)
            self.starting_time_period = j.get('starting_time_period', None)
            self.player_start_position = j.get('player_start_position', None)
            self.camera_start_position = j.get('camera_start_position', None)

            if is_resetting:
                if self.starting_time_period is not None:
                    self.gamestate.time_period = self.starting_time_period
                if self.player_start_position is not None:
                    self.gamestate.player.position = self.player_start_position
                if self.camera_start_position is not None:
                    self.gamestate.camera.position = self.camera_start_position

            for t in j['background']:
                self._load_tile(t, 0)
            for t in j['obstacles']:
                self._load_tile(t, 1)

    def save(self):
        with open(self.get_path(), 'w') as f:
            json.dump(self.as_dict(), f, indent=4, sort_keys=True)

    def update(self, dt):
        for tile in list(self.background.values()):
            tile.update(dt)
        for tile in list(self.obstacles.values()):
            old_pos = tile.position
            tile.update(dt)
            if old_pos != tile.position:
                self.remove_tile(old_pos, 1)
                self.add_tile(tile, 1)
