import os
import random

import pygame

from .baseobjects import Drawable
from .chapter_transition import ChapterTransition

class TileSet(object):
    def __init__(self, name):
        path = os.path.join('res', name + '.png')
        self.image = pygame.image.load(path).convert_alpha()
        self.tile_count = self.image.get_width() // Tile.SIZE


class Tile(Drawable):
    SIZE = 32

    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Tile, self).__init__(position, gamestate)
        self.tileset_name = tileset_name
        self.tileset_tile_idx = tileset_idx % TILESETS[tileset_name].tile_count
        self.behavior = 'tile'
        self.animating = False
        self.layer_index = layer_index
        self._anim_frame_timer = 0

    def draw_on(self, surface, cam):
        x = self.position[0] - cam.position[0]
        y = self.position[1] - cam.position[1]

        surface.blit(
            TILESETS[self.tileset_name].image,
            (x, y),
            (self.tileset_tile_idx * self.SIZE, self.gamestate.time_period * self.SIZE, self.SIZE, self.SIZE)
        )

        if self.animating:
            period = self.gamestate.fps * 0.1
            self._anim_frame_timer = (self._anim_frame_timer + 1) % period
            if self._anim_frame_timer == 0:
                self.tileset_tile_idx = (self.tileset_tile_idx + 1) % TILESETS[self.tileset_name].tile_count

    def get_rect(self):
        return pygame.Rect(
            self.position[0],
            self.position[1],
            self.SIZE,
            self.SIZE
        )

    def as_dict(self):
        return {
            'tileset_name': self.tileset_name,
            'tileset_idx': self.tileset_tile_idx,
            'position': self.position,
            'behavior': self.behavior
        }

    def remove(self):
        self.gamestate.room.remove_tile(self.position, self.layer_index)


class Movable(Tile):
    MOVE_SPEED = 0.1

    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Movable, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.movement_vector = None
        self.dest_position = None

    def is_moving(self):
        return self.movement_vector is not None

    def _check_collision(self, dest, movement_vector, tiles):
        can_pass = True
        if dest in tiles:
            tile = tiles[dest]
            if tile is not self:     
                #self.on_collide(tile, movement_vector)
                can_pass = tile.on_collide(self, movement_vector)


        for dx in range(-1, 2):
            for dy in range(-1, 2):
                x = dest[0] + Tile.SIZE * dx
                y = dest[1] + Tile.SIZE * dy

                if not (x,y) in tiles:
                    continue

                tile = tiles[(x,y)]

                if tile is not self:
                    if not tile.on_collide33(self, movement_vector):
                        can_pass = False
                    #self.on_collide33(self.movement_vector)

        return can_pass

    def move_direction(self, movement_vector):
        if not self.is_moving():
            # Move to next tile in direction specified
            dest = (
                self.position[0] + (movement_vector[0] * self.SIZE),
                self.position[1] + (movement_vector[1] * self.SIZE)
            )

            if not self._check_collision(dest, movement_vector, self.gamestate.room.obstacles) or \
                not self._check_collision(dest, movement_vector, self.gamestate.room.background):
                return

            self.movement_vector = movement_vector
            self.dest_position = dest

    @classmethod
    def _reached_pos(cls, curr, dest, direction):
        return (curr >= dest and direction >= 0) or \
                (curr <= dest and direction <= 0)

    def _move_toward_dest(self, speed):
        new_x = self.position[0] + (self.movement_vector[0] * speed)
        new_y = self.position[1] + (self.movement_vector[1] * speed)

        if self._reached_pos(new_x, self.dest_position[0], self.movement_vector[0]):
            new_x = self.dest_position[0]
        if self._reached_pos(new_y, self.dest_position[1], self.movement_vector[1]):
            new_y = self.dest_position[1]

        self.position = (new_x, new_y)

        if self.position == self.dest_position:
            self.movement_vector = None
            self.dest_position = None

    def update(self, dt):
        super(Movable, self).update(dt)

        if self.is_moving():
            speed = self.MOVE_SPEED * dt
            if self.gamestate.is_hyper:
                speed *= 2
            if self.gamestate.is_drunk:
                speed /= 2
            self._move_toward_dest(speed)


class Wall(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Wall, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'wall'

    def on_collide(self, other, movement_vector):
        # Solid
        return False

class Door(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0,
                 next_room_name='', next_room_player_pos=None, next_room_camera_pos=None,
                 chapter_transition=''):
        super(Door, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'door'
        self.next_room_name = next_room_name
        self.next_room_player_pos = next_room_player_pos
        self.next_room_camera_pos = next_room_camera_pos
        self.chapter_transition = chapter_transition
        self.hit_player = False

    def draw_on(self, surface, cam):
        if not self.gamestate.editor_editing:
            super(Door, self).draw_on(surface, cam)
        else:
            x = self.position[0] - cam.position[0]
            y = self.position[1] - cam.position[1]
            pygame.draw.rect(
                surface,
                (255, 0, 0),
                (x, y, Tile.SIZE, Tile.SIZE)
            )

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        self.hit_player = True
        return True

    def _should_transition_level(self):
        return self.hit_player and \
                self.position == self.gamestate.player.position

    def update(self, dt):
        super(Door, self).update(dt)
        if self._should_transition_level():
            if self.next_room_player_pos is not None:
                self.gamestate.player.position = self.next_room_player_pos

            if self.next_room_camera_pos is not None:
                self.gamestate.camera.position = self.next_room_camera_pos
            else:
                self.gamestate.camera.position = (0, 0)

            self.gamestate.load_room(self.next_room_name, bool(self.chapter_transition))
            self.hit_player = False
            self.gamestate.player.movement_vector = None
            self.gamestate.player.dest_position = None

            if self.chapter_transition:
                self.gamestate.transition_queue.insert(
                    0,
                    ChapterTransition(self.gamestate, self.chapter_transition)
                )

    def as_dict(self):
        d = super(Door, self).as_dict()
        d['next_room_name'] = self.next_room_name
        d['chapter_transition'] = self.chapter_transition

        if self.next_room_player_pos is not None:
            d['next_room_player_pos'] = self.next_room_player_pos
        else:
            d['next_room_player_pos'] = (
                (self.gamestate.screen_size[0] - Tile.SIZE) - self.position[0],
                (self.gamestate.screen_size[1] - Tile.SIZE) - self.position[1]
            )

        if self.next_room_camera_pos is not None:
            d['next_room_camera_pos'] = self.next_room_camera_pos
        else:
            d['next_room_camera_pos'] = (0, 0)

        return d

class AgeUpDoor(Door):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0,
                 next_room_name='', next_room_player_pos=None, next_room_camera_pos=None,
                 chapter_transition=''):
        super(AgeUpDoor, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx,
                                        next_room_name, next_room_player_pos, next_room_camera_pos, chapter_transition)
        self.behavior = 'ageupdoor'

    def draw_on(self, surface, cam):
        # Draw hint carpet
        self.gamestate.time_period = (self.gamestate.time_period + 1) % 3
        super(AgeUpDoor, self).draw_on(surface, cam)
        self.gamestate.time_period = (self.gamestate.time_period - 1) % 3

    def update(self, dt):
        if self._should_transition_level():
            self.gamestate.offset_time_period(1)
        super(AgeUpDoor, self).update(dt)

class AgeDownDoor(Door):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0,
                 next_room_name='', next_room_player_pos=None, next_room_camera_pos=None,
                 chapter_transition=''):
        super(AgeDownDoor, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx,
                                          next_room_name, next_room_player_pos, next_room_camera_pos, chapter_transition)
        self.behavior = 'agedowndoor'

    def draw_on(self, surface, cam):
        # Draw hint carpet
        self.gamestate.time_period = (self.gamestate.time_period - 1) % 3
        super(AgeDownDoor, self).draw_on(surface, cam)
        self.gamestate.time_period = (self.gamestate.time_period + 1) % 3

    def update(self, dt):
        if self._should_transition_level():
            self.gamestate.offset_time_period(-1)
        super(AgeDownDoor, self).update(dt)

class Knife(Door):
    KNIFE_QUIPS = [
        'Knives aren\'t toys!',
        'Well what did you think was going to happen?',
        'Goodbye cruel world!',
        'Mom said not to touch that!'
    ]

    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0,
                 next_room_name='', next_room_player_pos=None, next_room_camera_pos=None,
                 chapter_transition=''):
        super(Knife, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx,
                                    gamestate.last_chapter_start_name, next_room_player_pos, next_room_camera_pos,
                                    random.choice(self.KNIFE_QUIPS))
        self.behavior = 'knife'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            return super(Knife, self).on_collide(other, movement_vector)
        if self.gamestate.time_period == 1:
            self.remove()
            return True
        if self.gamestate.time_period == 2:
            return False
        return False


class MomChair(Movable):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(MomChair, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'mom_chair'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            return True
        if self.gamestate.time_period == 1:
            return False
        if self.gamestate.time_period == 2:
            self.move_direction(movement_vector)
            return False
        return False

class Poop(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Poop, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'poop'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            return True
        if self.gamestate.time_period == 1:
            return False
        if self.gamestate.time_period == 2:
            return False
        return False

class Tax(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Tax, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'tax'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            self.remove()
            return True
        if self.gamestate.time_period == 1:
            return False
        if self.gamestate.time_period == 2:
            self.gamestate.is_tax = True
            self.remove()
            return True
        return False

class Fireplace(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Fireplace, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'fireplace'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            return False
        if self.gamestate.time_period == 1:
            return False
        if self.gamestate.time_period == 2 and self.gamestate.is_bucket:
            self.gamestate.room.add_tile(Tile(self.tileset_name, self.position, self.layer_index, self.gamestate, 14), self.layer_index)
            return True
        return False

    def on_collide33(self, other, movement_vector):
        if not other.is_player():
            return True

        if self.gamestate.time_period == 0:
            return False
        return True

class Car(Movable):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Car, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'car'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            self.move_direction(movement_vector)
            return False
        if self.gamestate.time_period == 1:
            return True
        if self.gamestate.time_period == 2:
            return False
        return False

class Cookie(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Cookie, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'cookie'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            self.remove()
            self.gamestate.is_hyper = True
            return True
        if self.gamestate.time_period == 1:
            self.remove()
            return True
        if self.gamestate.time_period == 2:
            return False
        return False

class Drink(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Drink, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'drink'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        self.remove()
        if self.gamestate.time_period == 0:
            self.gamestate.is_pee = True
            return True
        if self.gamestate.time_period == 1:
            self.gamestate.is_hyper = True
            return True
        if self.gamestate.time_period == 2:
            self.gamestate.is_drunk = True
            return True
        return False

class Table(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Table, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'table'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            return False
        if self.gamestate.time_period == 1:
            self.remove()
            return True
        if self.gamestate.time_period == 2:
            return False
        return False

class Dumbbell(Movable):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Dumbbell, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'dumbbell'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            return False
        if self.gamestate.time_period == 1:
            return False
        if self.gamestate.time_period == 2:
            if self.gamestate.is_hyper:
                self.move_direction(movement_vector)
            return False
        return False

class Bed(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Bed, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'bed'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.is_drunk:
            return True
        return False

class TaxSign(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(TaxSign, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'tax_sign'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.is_tax:
            return True
        return False

class FullBucket(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(FullBucket, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'full_bucket'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            return False
        if self.gamestate.time_period == 1:
            return False
        if self.gamestate.time_period == 2:
            self.remove()
            self.gamestate.is_bucket = True
            return True
        return False

class Bucket(Movable):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(Bucket, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'bucket'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        if self.gamestate.time_period == 0:
            if self.gamestate.is_pee:
                self.gamestate.room.add_tile(FullBucket(self.tileset_name, self.position, self.layer_index, self.gamestate, 11), self.layer_index)
            return False
        if self.gamestate.time_period == 1:
            self.move_direction(movement_vector)
            return False
        if self.gamestate.time_period == 2:
            return True
        return False

class EndGame(Tile):
    def __init__(self, tileset_name, position, layer_index, gamestate, tileset_idx=0):
        super(EndGame, self).__init__(tileset_name, position, layer_index, gamestate, tileset_idx)
        self.behavior = 'endgame'

    def on_collide(self, other, movement_vector):
        if not other.is_player():
            return False

        self.gamestate.transition_queue.insert(0, ChapterTransition(self.gamestate, 'Home', 5))
        self.gamestate.transition_queue.insert(0, ChapterTransition(self.gamestate, '', 1))
        self.gamestate.transition_queue.insert(0, ChapterTransition(self.gamestate, 'By Matt Penny and Matt Diener', 5))
        self.gamestate.transition_queue.insert(0, ChapterTransition(self.gamestate, 'Thanks for playing', 0))

TILESETS = { name: TileSet(name) for name in ['floors', 'walls', 'walkdown', 'walkup', 'walkright', 'walkleft', 'obstacles', 'inventory'] }

TILE_BEHAVIORS = {
    'tile': Tile,
    'wall': Wall,
    'door': Door,
    'mom_chair': MomChair,
    'poop': Poop,
    'tax': Tax,
    'ageupdoor': AgeUpDoor,
    'agedowndoor': AgeDownDoor,
    'fireplace': Fireplace,
    'car': Car,
    'cookie': Cookie,
    'drink': Drink,
    'table': Table,
    'dumbbell': Dumbbell,
    'bed': Bed,
    'tax_sign': TaxSign,
    'full_bucket': FullBucket,
    'bucket': Bucket,
    'knife': Knife,
    'endgame': EndGame
}
